#!/usr/bin/env python3

import ast
import base64
import json
import logging
import psycopg2
import re
import requests

from decouple import Config, RepositoryEnv
from func_timeout import func_timeout, FunctionTimedOut
from netaddr import IPAddress
from pymemcache.client.base import Client
from time import sleep

##
# URIs
##

# Remote server URI
auth_uri = 'https://nemesida-security.com/nw/auth/v3'
rl_desc_uri = 'https://nemesida-security.com/nw/rl-desc/'
geoip_api_uri = 'http://geoip.nemesida-security.com/geoip/v2?ip='

# Local server URI
base_api_uri = '/nw-api'

# Attacks, scans URLs
set_attack_uri = base_api_uri + '/attack'
get_attack_uri = base_api_uri + '/get_attack'
get_attack_stats_uri = base_api_uri + '/get_attack_stats'
set_scanner_uri = base_api_uri + '/scan'

# Misc.
headers = {'User-agent': 'Nemesida WAF API'}
waf_id_ce = 'COMMUNITYEDITION'

##
# Log settings
##

logf = '/var/log/uwsgi/nw-api/nw-api-logging.log'
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
fh = logging.FileHandler(logf)
formatter = logging.Formatter('[%(asctime)s] %(levelname)s - %(message)s', '%m-%d %H:%M:%S')
fh.setFormatter(formatter)
log.addHandler(fh)

##
# Export settings
##

# init
proxy = {}
memcached_host = '127.0.0.1'
memcached_port = '11211'
settings_file = '/var/www/nw-api/settings.py'

try:

    # init
    env_config = Config(RepositoryEnv(settings_file))

    # extract values
    proxy_settings = env_config.get('PROXY', default='')
    db_host = env_config.get('DB_HOST', default='127.0.0.1')
    db_port = env_config.get('DB_PORT', default='5432')
    db_name = env_config.get('DB_NAME', default='waf')
    db_user = env_config.get('DB_USER', default='nw_api')
    db_pass = env_config.get('DB_PASS', default='')
    ro_mode = env_config.get('RO_MODE', default=False, cast=bool)

    memcached_host = env_config.get('MEMCACHED_HOST', default='127.0.0.1')
    memcached_port = env_config.get('MEMCACHED_PORT', default='11211')

    # DB settings processing
    if not db_pass:
        log.warning(
            'Option "db_pass" is not set, make changes to the configuration file and restart the services'
        )
        exit()

    # Proxy settings processing
    if proxy_settings:
        try:

            # init
            proxy = proxy_settings

            # set schema if not exists
            if not re.match(r'^https?://', proxy_settings.lower()):
                proxy = 'http://' + proxy_settings

            # logging
            log.info('Proxy server used: {}'.format(proxy))

            # set proxy
            proxy = {'http': proxy, 'https': proxy}

        except Exception as error:
            log.error(
                'An error occurred while adding the schema for proxy-server: {}'.format(error)
            )

except Exception as error:
    log.error(
        'An error occurred while exporting settings from file {}: {}'
        .format(settings_file, error)
    )
    exit()


##
# Memcached settings
##

memcached_setting = (memcached_host, int(memcached_port))

##
# Functions
##


def get_auth(key, agent_ip):

    # init
    waf_id = ''
    tarif_id = 0
    models_num = 0
    geo_token = ''

    # init memcached
    client = Client(memcached_setting)
    memc_key = 'auth_{}'.format(key)

    ##
    # Search data in memcached
    ##

    try:

        # get data from memcached
        memc_bin = client.get(memc_key)

        # if data in memcached
        if memc_bin:

            # processing memcached data
            data = memc_bin.decode('UTF-8')
            data = json.loads(data)

            # If lic. key is valid
            if data['status'] == 'success':

                # extract data
                waf_id = str(data['waf_id'])
                tarif_id = data['tarif_id']
                models_num = data['models_num']
                geo_token = data['geo_token']

        # return
        if waf_id:
            return waf_id, tarif_id, models_num, geo_token

    except Exception as e:
        log.error('An error occurred while processing WAF ID from the memcached: {}'.format(e))

    ##
    # If WAF ID not in memcached
    # receiving from the Nemesida WAF Auth
    ##

    while True:
        try:

            # init
            s = init_session()

            # send request
            result = s.post(
                auth_uri,
                headers=headers,
                proxies=proxy,
                data={'key': key},
                verify=True,
                timeout=15
            )
            status_code = result.status_code

            # processing
            if status_code == 201:
                try:

                    # init
                    data = json.loads(result.text)

                    # extract data
                    waf_id = str(data['waf_id'])
                    tarif_id = data['tarif_id']
                    models_num = data['models_num']
                    geo_token = data['geo_token']

                    # update WAF ID in memcached
                    try:
                        if waf_id:
                            client.set(memc_key, json.dumps(data), 600)
                    except Exception as e:
                        log.error(
                            'An error occurred while storing auth data for product key {}'
                            ' to the memcached from client {}: {}'
                            .format(key, agent_ip, e)
                        )

                    # return
                    return waf_id, tarif_id, models_num, geo_token

                except Exception as e:
                    log.error(
                        'An error occurred while processing auth data for product key {} from client {}: {}'
                        .format(key, agent_ip, e)
                    )
                    sleep(1)

            elif status_code == 401:
                log_key = 'x' * 26 + key[-6:]
                log.warning('Invalid product key {} received from client {}'.format(log_key, agent_ip))
                return waf_id, tarif_id, models_num, geo_token

        except Exception as e:
            log.error('An error occurred while processing the request from client {}: {}'.format(agent_ip, e))


def get_waf_id(key, agent_ip):
    try:
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)
        return waf_id
    except Exception as e:
        log.error('An error occurred while processing WAF ID: {}'.format(e))
        return False


def db_conn():
    try:

        connect = psycopg2.connect(
            database=db_name,
            user=db_user,
            password=db_pass,
            host=db_host,
            port=db_port,
            connect_timeout=10
        )

        return connect, connect.cursor()

    except Exception as e:
        log.error('An error occurred while connecting to the database: {}'.format(e))
        return False, False


def init_db():

    ##
    # Create new column if not exist
    ##

    try:

        log.info('The database is being initialized')

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            log.error('An error occurred while initializing the database: unable to connect to the database')
        else:

            """
            Create new tables if not exists
            """

            fname = 'api.sql'
            log.info('Database processing using {}'.format(fname))
            cur.execute(open(fname, 'r').read())
            connect.commit()
            log.info('The database has been processed successfully using {}'.format(fname))

            ##
            # Updating the tables
            ##

            log.info('Updating the database tables')

            """
            Table: attack
            """

            # Receive columns name
            query = """
                SELECT
                    column_name
                FROM
                    INFORMATION_SCHEMA.COLUMNS
                WHERE
                    TABLE_NAME = 'attack'
                """
            cur.execute(query,)
            res = cur.fetchall()

            # rename content to body
            if res and ('content',) in res:
                query = """ ALTER TABLE attack RENAME COLUMN content TO body """
                cur.execute(query,)
                connect.commit()

            # rename uri to url
            if res and ('uri',) in res:
                query = """ ALTER TABLE attack RENAME COLUMN uri TO url """
                cur.execute(query,)
                connect.commit()

            # add/drop/alter columns
            query = """
                ALTER TABLE
                    attack
                DROP COLUMN IF EXISTS
                    bot,
                ADD COLUMN IF NOT EXISTS
                    group_id text,
                ADD COLUMN IF NOT EXISTS
                    blocked boolean,
                ADD COLUMN IF NOT EXISTS
                    possible boolean,
                ADD COLUMN IF NOT EXISTS
                    cc char(2) DEFAULT NULL,
                ALTER COLUMN
                    url DROP NOT NULL,
                ALTER COLUMN
                    param DROP NOT NULL,
                ALTER COLUMN
                    other_headers DROP NOT NULL,
                ALTER COLUMN
                    referer DROP NOT NULL,
                ALTER COLUMN
                    mz DROP NOT NULL,
                ALTER COLUMN
                    description DROP NOT NULL,
                ALTER COLUMN
                    rule_id DROP NOT NULL,
                ALTER COLUMN
                    rule_id SET DEFAULT 0
            """
            cur.execute(query,)
            connect.commit()

            # rename sequence
            query = """ ALTER SEQUENCE IF EXISTS attack_id_seq RENAME TO attack_ids """
            cur.execute(query,)
            connect.commit()

            # drop indexes
            query = """ DROP INDEX IF EXISTS idx_mtp """
            cur.execute(query,)
            connect.commit()

            log.info('Table "attack" has been successfully updated or no migration is required')

            """
            Table: ml
            """

            # drop columns
            query = """ ALTER TABLE ml DROP COLUMN IF EXISTS ml_status """
            cur.execute(query,)
            connect.commit()

            log.info('Table "ml" has been successfully updated or no migration is required')

            """
            Table: scan_report
            """

            ##
            # Add new column mz
            ##

            query = """
                ALTER TABLE
                    scan_report
                ADD COLUMN IF NOT EXISTS
                    mz text
            """
            cur.execute(query,)
            connect.commit()

            # drop NOT NULL for specific columns
            query = """
                ALTER TABLE
                    scan_report
                ALTER COLUMN
                    param DROP NOT NULL,
                ALTER COLUMN
                    payload DROP NOT NULL,
                ALTER COLUMN
                    type DROP NOT NULL,
                ALTER COLUMN
                    url DROP NOT NULL,
                ALTER COLUMN
                    data DROP NOT NULL
            """
            cur.execute(query,)
            connect.commit()

            # rename sequence
            query = """
                ALTER SEQUENCE IF EXISTS
                    scan_report_id_seq
                RENAME TO
                    scan_report_ids
            """
            cur.execute(query,)
            connect.commit()

            log.info('Table "scan_report" has been successfully updated or no migration is required')

            """
            Table: dyn_settings
            """

            # add new column nwaf_b64_decode_disable
            query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_b64_decode_disable text """
            cur.execute(query,)
            connect.commit()

            # add new column nwaf_geoip_mla_disable
            query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_geoip_mla_disable text """
            cur.execute(query,)
            connect.commit()

            # add new column nwaf_rfc_violation_wl
            query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_rfc_violation_wl text """
            cur.execute(query,)
            connect.commit()
            
            # add new column nwaf_rfc_violation_lm
            query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_rfc_violation_lm text """
            cur.execute(query,)
            connect.commit()

            log.info('Table "dyn_settings" has been successfully updated or no migration is required')

            """
            Table: openapi_url
            """
            
            # add/drop/alter columns
            query = """
                ALTER TABLE
                    openapi_url
                ALTER COLUMN
                    timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
            """
            cur.execute(query,)
            connect.commit()

            log.info('Table "openapi_url" has been successfully updated or no migration is required')

            """
            Table: openapi_schemas
            """
            
            # add/drop/alter columns
            query = """
                ALTER TABLE
                    openapi_schemas
                ALTER COLUMN
                    timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
            """
            cur.execute(query,)
            connect.commit()

            log.info('Table "openapi_schemas" has been successfully updated or no migration is required')

            """
            Table: openapi_schema_paths
            """
            
            # add/drop/alter columns
            query = """
                ALTER TABLE
                    openapi_schema_paths
                ALTER COLUMN
                    timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
            """
            cur.execute(query,)
            connect.commit()

            log.info('Table "openapi_schemas_path" has been successfully updated or no migration is required')

            """
            Table: openapi_exclude_paths
            """
            
            # add/drop/alter columns
            query = """
                ALTER TABLE
                    openapi_exclude_paths
                ALTER COLUMN
                    timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
            """
            cur.execute(query,)
            connect.commit()

            log.info('Table "openapi_exclude_paths" has been successfully updated or no migration is required')

            ##

            connect.close()
            log.info('The database initialization is complete')

    except Exception as e:
        log.error('An error occurred while updating the table structure: {}'.format(e))


def init_session():

    s = requests.Session()
    s.trust_env = False
    return s


def get_rldesc(rule_id):

    # init
    rldesc = ''

    try:

        # logging
        log.info('Updating description for RuleID {}'.format(rule_id))

        # init
        s = init_session()
        data = {'rule_id': rule_id}
        data = json.dumps(data)

        # receiving rule description
        state = s.post(
            rl_desc_uri,
            proxies=proxy,
            headers=headers,
            data=data,
            timeout=120
        )
        status_code = state.status_code

        # processing result
        if status_code != 200 and status_code != 201:
            log.error(
                'Incorrect server answer for RuleID {}: {}, text: {}'.format(rule_id, state.status_code, state.text)
            )
            return rldesc

        desc_dict = state.text
        desc_dict = json.loads(desc_dict)

        return desc_dict['desc']

    except Exception as e:
        log.error(
            'An error occurred in while processing description for rule {}: {}'.format(rule_id, e)
        )
        return rldesc


def get_geoip(ip, waf_id):
    try:
        
        # remove spaces
        ip = ip.strip()

        # init memcached
        client = Client(memcached_setting)
        memc_key = 'geoip_v3_' + str(ip)

        # skip for non-public IP
        try:
            if IPAddress(ip).is_private() or IPAddress(ip).is_loopback() or IPAddress(ip).is_multicast():
                return 'Private IP'
        except Exception:
            pass

        try:

            # search data in memcached
            memout = client.get(memc_key)

            # if data in memcached
            if memout:

                # load as a dictionary
                data = ast.literal_eval(memout.decode('UTF-8', 'ignore'))

                # processing success
                if data and 'status' in data and data['status'] == 'success':
                    return data

        except Exception as e:
            log.error('An error occurred while processing GeoIP data from memcached: {}'.format(e))
        
        # skip empty/CE WAF ID
        if not waf_id.strip() or waf_id == waf_id_ce:
            return False
        
        # get GeoToken
        geo_token = get_geo_token(waf_id)
        if not geo_token:
            return False

        ##
        # Receive GeoIP from the remote server
        # if data not found in the memcached and IP is a public
        ##

        # init
        s = init_session()

        # receiving GeoIP
        result = s.get(
            geoip_api_uri + ip + '&' + geo_token,
            proxies=proxy,
            headers=headers,
            verify=True,
            timeout=5,
        )

        # processing status code
        if not re.match(r'^20', str(result.status_code)):
            log.error(
                'An error occurred while processing data from GeoIP server: {}'
                .format(result)
            )
            return False

        ##
        # JSON processing
        ##

        # load to JSON
        data = json.loads(result.text)

        # processing private
        if data and 'status' in data and data['status'] == 'fail':
            if 'description' not in data:
                log.error(
                    'An error occurred while processing JSON data from GeoIP server for IP {}: no description in {}'
                    .format(ip, data)
                )
            elif data['description'] == 'Private IP':
                return 'Private IP'

        # processing success
        elif data and 'status' in data and data['status'] == 'success' and data['cc']:
            try:
                # create dictionary with certain keys
                geoip_extract_list = ['cc', 'mobile', 'proxy', 'hosting', 'crawler', 'status']
                data = {k: v for k, v in data.items() if k in geoip_extract_list}
                # add GeoIP to memcached and return data
                client.set(memc_key, data, 86400)  # 86400 - 1 d.
                return data
            except Exception as e:
                log.error(
                    'An error occurred while storing GeoIP data in memcached: {} (data: {})'
                    .format(e, data)
                )

    except Exception as e:
        log.error(
            'An error occurred while processing JSON data from GeoIP server for IP {}: {}'
            .format(ip, e)
        )

    # default action
    return False


def get_geoip_cc(ip, waf_id):
    try:

        # init
        resp_timeout = 2

        try:
            # get CC from remote server
            data = func_timeout(resp_timeout, get_geoip, args=(ip, waf_id,))
        except FunctionTimedOut:
            log.warning('A timeout occurred while receiving GeoIP data for {}'.format(ip))
            return None

        # processing private
        if data and data == 'Private IP':
            return ''

        # processing success
        if data and 'status' in data and data['status'] == 'success' and data['cc']:
            return data['cc']

    except Exception as e:
        log.error(
            'An error occurred while processing CountryCode for {}: {}'
            .format(ip, e)
        )

    # default response
    return None


def geo_token_update(key, agent_ip):
    try:
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)
        put_geo_token_to_db(waf_id, geo_token)
    except Exception as e:
        log.error('An error occurred while processing GeoToken: {}'.format(e))


def get_geo_token(waf_id):
    try:

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            return False

        # search GeoToken in the database        
        query = """ SELECT geo_token FROM geo_tokens WHERE waf_id = %s """
        cur.execute(query, (waf_id,))
        row = cur.fetchone()

        # processing result
        return row[0] if row else False

    except Exception as e:
        log.error(
            'An error occurred while processing GeoToken for WAF ID {} in the database: {}'.format(waf_id, e)
        )
        return False


def put_geo_token_to_db(waf_id, geo_token):
    try:

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            return False
        
        # store GeoToken to the database
        query = """ INSERT INTO geo_tokens (waf_id, geo_token) VALUES (%s, %s) ON CONFLICT DO NOTHING """
        cur.execute(query, (waf_id, geo_token,))
        connect.commit()

    except Exception as e:
        log.error('An error occurred during storing data to {}: {}'.format(geo_token, e))


def get_agent_ip(request):

    # extract client IP
    try:
        return request.remote_addr
    except Exception:
        return ''


def set_attack_processing(data, agent_ip):
    try:

        # init memcached
        client = Client(memcached_setting)

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            return

        ##
        # Processing dictionary of the attacks
        ##

        for item in data:
            try:

                # extract lic. key
                key = str(item['waf_lic_key'])

                # WAF ID processing
                if key == 'none':
                    waf_id = waf_id_ce
                else:
                    waf_id = get_waf_id(key, agent_ip)
                    geo_token_update(key, agent_ip)
                    if not waf_id:
                        continue

                # init
                fields_dict = {}

                # description fix
                if 'description' not in item:
                    item['description'] = ''

                # group_id fix
                if 'group_id' not in item:
                    item['group_id'] = ''

                # url fix
                if 'uri' in item:
                    item['url'] = item['uri']
                    item.pop('uri')

                # rename content to boy
                try:
                    if 'content' in item:
                        item['body'] = item.pop('content')
                except Exception as e:
                    log.error(
                        'An error occurred while renaming column "content" to "body": {}'.format(e)
                    )

                ##
                # b64decode
                ##

                # init
                b64_fields = (
                    'ip',
                    'request_id',
                    'vhost',
                    'url',
                    'param',
                    'mz',
                    'method',
                    'ua',
                    'referer',
                    'other_headers',
                    'body',
                    'cookie',
                    'description',
                    'group_id',
                )

                # decode custom field
                for field in b64_fields:
                    try:
                        # base64 decode
                        fields_dict[field] = base64.b64decode(item[field]).decode('UTF-8', 'ignore')
                    except Exception as e:
                        log.error(
                            'An error occurred while decoding the field (client IP: {}): {}'
                            .format(agent_ip, e)
                        )
                        fields_dict[field] = 'API DECODE ERROR'

                ##
                # Extract data from the JSON item
                ##

                # extract request ID
                try:
                    request_id = str(fields_dict['request_id'])
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "request_id" (client IP: {}): {}'
                        .format(agent_ip, e)
                    )
                    request_id = '0'

                # extract vhost
                try:
                    vhost = str(fields_dict['vhost'])
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "vhost" from request {}: {}'
                        .format(request_id, e)
                    )
                    vhost = ''

                # extract rule ID
                try:
                    rule_id = item['rule_id']
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "rule_id" from request {}: {}'
                        .format(request_id, e)
                    )
                    rule_id = 0

                # extract bt
                try:
                    bt = item['bt']
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "bt" from request {}: {}'.format(request_id, e)
                    )
                    bt = 0

                # extract ip
                try:
                    ip = str(fields_dict['ip'])
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "ip" from request {}: {}'.format(request_id, e)
                    )
                    ip = ''

                # extract mz
                try:

                    # convert to list
                    mz_list = fields_dict['mz'].split('|')

                    # mz normalization
                    mz_list = attack_zone_replace(mz_list)

                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "mz" from request {}: {}'.format(request_id, e)
                    )
                    mz_list = []

                # extract body
                try:
                    body = bytearray(fields_dict['body'], 'UTF-8')
                except Exception as e:
                    log.error(
                        'An error occurred while extracting field "body" from request {}: {}'.format(request_id, e)
                    )
                    body = ''

                # Processing CountryCode
                cc = get_geoip_cc(ip, waf_id)

                ##
                # Processing for each zone in mz
                ##

                for mz in mz_list:
                    try:

                        # init
                        description = ''

                        # processing specific bt
                        if (bt == '1' or bt == '2') and rule_id != '0':

                            ##
                            # Receive description
                            ##

                            # init
                            memc_key = 'rldsc_' + rule_id

                            # Search description in the memcached
                            try:

                                # get data from memcached
                                memc_bin = client.get(memc_key)

                                # if data exist in memcached
                                if memc_bin:
                                    description = memc_bin.decode('UTF-8')

                            except Exception as e:
                                log.error(
                                    'An error occurred while processing rule description from the memcached: {}'
                                    .format(e)
                                )
                                description = ''

                            # Receive the description from the remote server if description is empty
                            if not description or description == '-':

                                # receive the description
                                description = get_rldesc(rule_id)

                                # update data in memcached
                                try:
                                    if description:
                                        client.set(memc_key, description, 86400)  # 84600 - 1 day
                                except Exception as e:
                                    log.error(
                                        'An error occurred while storing description "{}" for rule ID {}'
                                        ' in the memcached: {}'
                                        .format(description, rule_id, e)
                                    )

                        elif bt == '4':
                            description = 'Virus (' + fields_dict['description'] + ')'

                        elif bt == '5':
                            mz = ''
                            description = fields_dict['description']

                        elif bt == '7':
                            description = fields_dict['description']
                            if not description or description == '-':
                                description = 'Brute-force'

                        elif bt == '9':
                            description = fields_dict['description']
                            if not description or description == '-':
                                description = 'Flood'

                        elif bt == '10':
                            mz = ''
                            description = fields_dict['description']

                        else:
                            description = fields_dict['description']
                            if not description:
                                description = ''

                        ##
                        # Store attack to the database
                        ##

                        try:

                            # preparing a query
                            query = """
                                INSERT INTO attack (
                                    timestamp, lm, bt, mz, ip, vhost, rule_id,
                                    request_id, description, waf_id, url, method,
                                    referer, param, ua, cookie, other_headers, body,
                                    group_id, blocked, possible,cc
                                ) VALUES (
                                    to_timestamp(%s), %s, %s, %s, %s, %s, %s,
                                    %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                                    %s, %s, %s, %s
                                ) ON CONFLICT DO NOTHING
                           """

                            ar = (
                                item.get('timestamp', ''),
                                item.get('lm', ''),
                                bt,
                                mz,
                                ip,
                                vhost,
                                rule_id,
                                request_id,
                                description,
                                waf_id,
                                fields_dict['url'],
                                fields_dict['method'],
                                fields_dict['referer'],
                                fields_dict['param'],
                                fields_dict['ua'],
                                fields_dict['cookie'],
                                fields_dict['other_headers'],
                                body,
                                fields_dict['group_id'],
                                item.get('blocked', None),
                                item.get('possible', None),
                                cc,
                            )

                            # log.info('Preparing the query for writing to the database: {}'.format(ar))

                            try:

                                # insert data to the DB
                                cur.execute(query, ar)
                                connect.commit()

                                # logging
                                log.info(
                                    'Request {} (BT: {}, client: {}, server: {}, WAF ID: {}) received from {}'
                                    .format(request_id, bt, ip, vhost, waf_id, agent_ip)
                                )

                            except Exception as e:
                                log.error(
                                    'An error occurred while connecting to the database'
                                    ' while storing request {} (BT: {}, client: {}, server: {}, WAF ID: {}): {}'
                                    .format(request_id, bt, ip, vhost, waf_id, e)
                                )

                        except Exception as e:
                            log.error(
                                'An error occurred during inserting request {}'
                                ' (BT: {}, client: {}, server: {}, WAF ID: {}) to the database: {}'
                                .format(request_id, bt, ip, vhost, waf_id, e)
                            )

                    except Exception as e:
                        log.error('An error occurred while processing zone {} in {}: {}'.format(mz, request_id, e))

            except Exception as e:
                log.error('An error occurred while processing JSON item: {}'.format(e))

        # closing connection to the DB
        connect.close()

    except Exception as e:
        log.error(
            'An error occurred while processing data in set_attack_processing: {}'
            .format(e)
        )


def attack_zone_replace(mz):

    # init
    ret_mz = []

    # replace conditions
    repl = {
        "uri": "URL",
        "args": "ARGS",
        "body": "BODY",
        "ua": "User-Agent",
        "cookie": "Cookie",
        "referer": "Referer",
        "other_headers": "HEADERS",
    }

    # processing for each zone in mz list
    for z in mz:

        # zone normalization
        for k, v in repl.items():
            z = z.replace(k, v)

        # update result list
        ret_mz.append(z)

    return ret_mz


def vhost_normalization(vhost):
    try:

        # init
        dot = r'^\.'
        sub = r'^\*\.'

        # *
        if vhost == '*':
            vhost = 'all'

        # .example.com
        elif re.search(dot, vhost):
            vhost = re.sub(dot, '', vhost)

        # *.example.com
        elif re.search(sub, vhost):
            vhost = re.sub(sub, 'sub_', vhost)

    except Exception as e:
        log.error('An error occurred while vhost normalization (vhost: {}): {}'.format(vhost, e))

    return vhost


def idna_vhost_list_normalization(vhost):
    try:

        # dot
        if vhost.startswith('.'):

            # processing for non-latin vhost
            try:
                vhost = '.' + vhost[1:].encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

        # other
        else:

            # processing for non-latin vhost
            try:
                vhost = vhost.encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

    except Exception as e:
        log.error('An error occurred while normalizing vhost {}: {}'.format(vhost, e))

    return vhost


def get_data_from_json(request, url):

    # extract client IP
    agent_ip = get_agent_ip(request)

    # load JSON
    try:
        return json.loads(request.data)
    except Exception as e:
        log.error(
            'An error occurred while processing the request to {} (client IP: {}): {}'
            .format(url, agent_ip, e)
        )
        return None


def is_base64(s):

    ##
    # Check is data in base64
    ##

    # data is not in Base64 if contains only numbers
    if re.match(r'^\d+={,2}$', s):
        return False

    # check is data in Base64
    try:
        return base64.b64encode(base64.b64decode(s.encode())).decode('UTF-8', 'ignore') in [s, s + '=', s + '==']
    except Exception:
        return False
