#!/usr/bin/env python3

from psycopg2.extras import RealDictCursor

from core import base_api_uri
from core import db_conn
from core import log

##
# URLs
##

set_mt_data_uri = base_api_uri + '/set_mt_data_uri'
get_mt_data_uri = base_api_uri + '/get_mt_data_uri'
get_mt_hash_uri = base_api_uri + '/get_mt_hash_uri'

##
# MT parameters
##

mt_data_keys = [
    'bt',
    'timestamp',
    'request_id',
    'mz',
    'ip',
    'vhost',
    'method',
    'uri',
    'args',
    'cookie',
    'user-agent',
    'referer',
    'body',
    'other_headers',
    'mz_origin'
]

mt_zone_map = {
    'body': 'Body',
    'uri': 'URL',
    'args': 'ARGS',
    'cookie': 'Cookie',
    'ua': 'UA',
    'referer': 'Refs.',
    'vhost': 'Host',
    'ip': 'IP',
    'other_headers': 'Headers',
}


##
# Functions
##

def lower_dict_keys(data_dict):
    return {key.lower() if type(key) == str else key: value for key, value in data_dict.items()}


def mt_get_export(waf_id):

    # init
    res = ()

    try:

        # init connection to the database
        connect, cur = db_conn()
        cur = connect.cursor(cursor_factory=RealDictCursor)
        if not connect or not cur:
            return False

        query = """ SELECT * FROM mt_export WHERE waf_id = %s """
        cur.execute(query, (waf_id,))
        res = cur.fetchall()
        connect.commit()
        connect.close()

    except Exception as e:
        log.error('An error occurred while processing data in mt_export for WAF ID {}: {}'.format(waf_id, e))

    return res


def mt_get_export_md5(waf_id):

    # init
    res = '0'

    try:

        # init connection to the database
        connect, cur = db_conn()
        cur = connect.cursor(cursor_factory=RealDictCursor)
        if not connect or not cur:
            return False

        query = """ SELECT md5(CAST(array_agg(mt_export.*) AS TEXT)) FROM mt_export WHERE waf_id = %s """
        cur.execute(query, (waf_id,))
        res = cur.fetchall()
        connect.commit()
        connect.close()

        if res and res[0]['md5']:
            res = str(res[0]['md5'])
        else:
            res = '0'

    except Exception as e:
        log.error('An error occurred while processing hash in mt_export for WAF ID {}: {}'.format(waf_id, e))

    return res


def mt_data_rename_fields(data):

    fields = (
        ('url', 'uri'),
        ('ua', 'user-agent'),
        ('headers', 'other_headers'),
    )

    if type(data) is dict:
        data = lower_dict_keys(data)
    if type(data) is list:
        data = [element.lower() for element in data]
    for src, dst in fields:
        if src in data:
            if type(data) is dict:
                data[dst] = data[src]
                data.pop(src)
            elif type(data) is list:
                data.append(dst)
                data.remove(src)

    return data


def mt_data_validation(data):
    missing_values = set(data.keys()) - set(mt_data_keys)
    if len(missing_values) > 0:
        return False
    unnecessary_values = set(data.keys()) ^ set(mt_data_keys)
    if len(unnecessary_values) > 0:
        return False
    return True
