#!/usr/bin/env python3

import ast
import base64
import hashlib
import OpenSSL.crypto
import re
import validators

from cryptography.hazmat.primitives.serialization import load_pem_parameters
from netaddr import IPNetwork, IPRange
from pymemcache.client.base import Client

from core import db_conn, log, memcached_setting, base_api_uri, is_base64


##
# URLs
##

get_nginx_status_uri = base_api_uri + '/nginx/get_status'

set_nginx_global_uri = base_api_uri + '/nginx/set_nginx'
get_nginx_global_uri = base_api_uri + '/nginx/get_nginx'
get_nginx_global_hash_uri = base_api_uri + '/nginx/get_nginx_hash'

set_nginx_vhost_uri = base_api_uri + '/nginx/set_vhost'
get_nginx_vhost_uri = base_api_uri + '/nginx/get_vhost'
get_nginx_vhost_hash_uri = base_api_uri + '/nginx/get_vhost_hash'

set_nginx_upstream_uri = base_api_uri + '/nginx/set_upstream'
get_nginx_upstream_uri = base_api_uri + '/nginx/get_upstream'
get_nginx_upstream_hash_uri = base_api_uri + '/nginx/get_upstream_hash'

set_nginx_certificate_uri = base_api_uri + '/nginx/set_certificate'
get_nginx_certificate_uri = base_api_uri + '/nginx/get_certificate'
get_nginx_certificate_hash_uri = base_api_uri + '/nginx/get_certificate_hash'

set_nginx_dhparam_uri = base_api_uri + '/nginx/set_dhparam'
get_nginx_dhparam_uri = base_api_uri + '/nginx/get_dhparam'
get_nginx_dhparam_hash_uri = base_api_uri + '/nginx/get_dhparam_hash'

get_nginx_with_active = [get_nginx_global_uri, get_nginx_vhost_uri]
get_nginx_with_format = [get_nginx_global_uri, get_nginx_vhost_uri, get_nginx_upstream_uri]

# Command
nginx_cmd = ['set', 'del']

##
# Columns
##

nginx_global_cols = [
    "active", "resolver", "resolver_timeout", "set_real_ip_from", "real_ip_header",
    "client_max_body_size", "limit_req_zone"
]

nginx_vhost_cols = [
    "active", "name", "listen", "server_name", "certificate_name", "dhparam_name", "ssl_session_timeout",
    "ssl_ciphers", "ssl_protocols", "root", "index", "locations"
]

nginx_upstream_cols = [
    "name", "balancing_method", "servers"
]

nginx_certificate_cols = [
    "name", "certificate", "certificate_key"
]

nginx_dhparam_cols = [
    "name", "dhparam"
]

##
# Keys
##

nginx_name = {
    'name': 'name', 'certificate_name': 'name', 'dhparam_name': 'name'
}

nginx_global = {
    'resolver': 'str', 'resolver_timeout': 'int', 'set_real_ip_from': 'list',
    'real_ip_header': 'str', 'client_max_body_size': 'int',
    'limit_req_zone': 'list'
}

nginx_vhost = {
    'active': 'bool', 'listen': 'str', 'server_name': 'str',
    'certificate_name': 'str', 'dhparam_name': 'str', 'ssl_session_timeout': 'int', 'ssl_ciphers': 'str',
    'ssl_protocols': 'str', 'root': 'str', 'index': 'str', 'locations': 'lod'
}

nginx_vhost_locations = {
    'location': 'str', 'pass_type': 'str', 'static_pass_param': 'dict',
    'proxy_pass_param': 'dict', 'memcached_pass_param': 'dict',
    'limit_req': "list", "limit_req_status": 'int'
}

nginx_vhost_locations_static_pass = {
    'path': 'str', 'index': 'str', 'autoindex': 'str'
}

nginx_vhost_locations_proxy_pass = {
    'proxy_pass': 'str', 'proxy_http_version': 'float', 'proxy_set_header': 'str', 'proxy_connect_timeout': 'int',
    'proxy_send_timeout': 'int', 'proxy_read_timeout': 'int', 'proxy_intercept_error': 'str'
}

nginx_vhost_locations_memcached_pass = {
    'set': 'str', 'memcached_pass': 'str', 'add_header': 'str', 'error_page': 'str'
}

nginx_upstream = {
    'balancing_method': 'str', 'servers': 'lod'
}

nginx_upstream_servers = {
    'server': 'str', 'weight': 'int', 'fail_timeout': 'int', 'max_fails': 'int', 'backup': 'bool', 'down': 'bool'
}

nginx_certificate = {
    'certificate': 'b64', 'certificate_key': 'b64'
}

nginx_dhparam = {
    'dhparam': 'b64'
}

##
# Join all keys with type
##

nginx_all = {}
nginx_all.update(nginx_name)
nginx_all.update(nginx_global)
nginx_all.update(nginx_vhost)
nginx_all.update(nginx_vhost_locations)
nginx_all.update(nginx_vhost_locations_static_pass)
nginx_all.update(nginx_vhost_locations_proxy_pass)
nginx_all.update(nginx_vhost_locations_memcached_pass)
nginx_all.update(nginx_upstream)
nginx_all.update(nginx_upstream_servers)
nginx_all.update(nginx_certificate)
nginx_all.update(nginx_dhparam)

nginx_all_keys = []
nginx_all_keys.extend(list(nginx_name.keys()))
nginx_all_keys.extend(list(nginx_global.keys()))
nginx_all_keys.extend(list(nginx_vhost.keys()))
nginx_all_keys.extend(list(nginx_vhost_locations.keys()))
nginx_all_keys.extend(list(nginx_vhost_locations_static_pass.keys()))
nginx_all_keys.extend(list(nginx_vhost_locations_proxy_pass.keys()))
nginx_all_keys.extend(list(nginx_vhost_locations_memcached_pass.keys()))
nginx_all_keys.extend(list(nginx_upstream.keys()))
nginx_all_keys.extend(list(nginx_upstream_servers.keys()))
nginx_all_keys.extend(list(nginx_certificate.keys()))
nginx_all_keys.extend(list(nginx_dhparam.keys()))

##
# List of types
##

nginx_format_types = ['str', 'b64', 'int', 'float', 'bool', 'list', 'dict', 'lod']

nginx_type_str = [k for k, v in nginx_all.items() if v == 'str']
nginx_type_b64 = [k for k, v in nginx_all.items() if v == 'b64']
nginx_type_int = [k for k, v in nginx_all.items() if v == 'int']
nginx_type_float = [k for k, v in nginx_all.items() if v == 'float']
nginx_type_bool = [k for k, v in nginx_all.items() if v == 'bool']
nginx_type_list = [k for k, v in nginx_all.items() if v == 'list']
nginx_type_dict = [k for k, v in nginx_all.items() if v == 'dict']
nginx_type_lod = [k for k, v in nginx_all.items() if v == 'lod']


def ip_validation(data):
    try:
        
        if '-' in data:  # processing IP range
            IPRange(data.split('-')[0], data.split('-')[1])
        else:
            IPNetwork(data)  # processing IPv4/IPv6/CIDR
        
        return True
    
    except Exception as e:
        log.error('An error occurred while IP validation for "{}": {}'.format(data, e))
        return False


def ip_domain_validation(data):
    
    # init
    if not isinstance(data, list):
        data = [data]

    # validation
    for item in data:

        # IP validation
        try:
            
            # processing IP range
            if '-' in item:
                IPRange(item.split('-')[0], item.split('-')[1])
            # processing IPv4/IPv6/CIDR
            else:
                IPNetwork(item)
            
            # validate the next item
            continue

        except Exception as e:
            err = str(e)

        # Domain validation if item is not an IP
        try:

            if item == 'localhost':
                continue
            elif validators.domain(item):
                continue

        except Exception as e:
            err = str(e)

        # if not validated
        log.error('An error occurred while IP/domain validation for "{}": {}'.format(item, err))
        return False

    # default action
    return True


def get_nginx_hash(table, waf_id):
    
    # init memcached
    client = Client(memcached_setting)
    memc_key = 'hash_' + table + '_' + waf_id

    try:

        # get data from memcached
        memc_bin = client.get(memc_key)
        if memc_bin:
            res = memc_bin.decode('UTF-8')
        else:
            set_nginx_hash(table, waf_id)
            memc_bin = client.get(memc_key)
            res = memc_bin.decode('UTF-8') if memc_bin else None

    except Exception as e:
        res = False
        log.error('An error occurred while extracting the hash for "{}" table: {}'.format(table, e))

    return res


def set_nginx_hash(table, waf_id):
    
    # init
    client = Client(memcached_setting)
    memc_key = 'hash_' + table + '_' + waf_id

    try:

        ##
        # Extract data from the database
        ##

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            err_msg = 'An error occurred while connection to the database'
            log.error(err_msg)

        # send a query
        query = """ SELECT * FROM {} WHERE waf_id = %s """.format(table)
        cur.execute(query, (waf_id,))
        res = cur.fetchall()
        connect.close()

        # Processing and hashing data
        if res:
            # hashing data
            hdata = hashlib.sha256(str(res).encode()).hexdigest()
            # set into memcached
            client.set(memc_key, hdata)
        else:
            # delete data from the memcached
            client.delete(memc_key)

    except Exception as e:
        err_msg = 'An error occurred while updating the hash from data in table "{}": {}'.format(table, e)
        log.error(err_msg)


def nginx_format_processing(nginx_type, data):

    # init
    res = ''
    indent_1 = ' ' * 2
    indent_2 = indent_1 * 2

    try:

        ##
        # Processing data
        ##

        if nginx_type == 'global':        
            
            # processing the dictionary
            for k, v in data.items():

                # value with size
                if k == 'client_max_body_size':
                    v = str(v) + 'm'

                # value as a list
                elif k in ['set_real_ip_from', 'limit_req_zone']:
                    for i in ast.literal_eval(v):
                        # if not empty
                        if i:
                            res = res + k + ' ' + str(i) + ';\n'                    
                    # skip updating the format below
                    continue
                
                # update format
                res = res + k + ' ' + str(v) + ';\n'

            # response
            return res

        elif nginx_type == 'vhost':
            
            # set header
            res = 'server {\n'

            # processing the dictionary
            for k, v in data.items():

                ##
                # Update value
                ##
                
                if k == 'certificate_name':
                    v = 'nwaf/global/nginx/ssl/crt/' + v + '.crt'
                
                elif k == 'certificate_key':
                    v = 'nwaf/global/nginx/ssl/key/' + v + '.key'
                
                elif k == 'dhparam_name':
                    v = 'nwaf/global/nginx/ssl/dh/' + v + '.pem'

                elif k == 'ssl_session_timeout':
                    v = str(v) + 's'

                elif k == 'locations':

                    # processing each dictionary in locations list
                    for i in ast.literal_eval(data['locations']):
                        
                        # extract location and pass_type
                        res = res + indent_1 + 'location ' + i['location'] + ' {\n'
                        
                        # processing locations items
                        for ik, iv in i.items():
                            if ik in ['proxy_connect_timeout', 'proxy_send_timeout', 'proxy_read_timeout']:
                                res = res + indent_2 + ik + ' ' + str(iv) + 's;\n'
                            else:
                                res = res + indent_2 + ik + ' ' + str(iv) + ';\n'

                    # skip updating the format below
                    continue

                # update format
                res = res + indent_1 + k + ' ' + v + ';\n'

            # set footer
            res = res + '}\n'

            # response
            return res

        elif nginx_type == 'upstream':
    
            # set header
            res = 'upstream ' + data['name'] + ' {\n'
            del data['name']

            # set balancing method
            if data['balancing_method']:
                res = res + indent_1 + data['balancing_method'] + ';\n'
                del data['balancing_method']
            
            # processing the dictionary
            for item in ast.literal_eval(data['servers']):

                # extract server
                server = item['server']
                del item['server']

                # set sub-header
                server_params = indent_1 + 'server' + ' ' + server

                # processing server's params
                for k, v in item.items():

                    if k in ['weight', 'max_fails']:
                        server_params = server_params + ' ' + k + '=' + str(v)
                    elif k in ['fail_timeout']:
                        server_params = server_params + ' ' + k + '=' + str(v) + 's'
                    elif k in ['backup', 'down']:
                        if v:
                            server_params = server_params + ' ' + k
                    else:
                        server_params = server_params + ' ' + k + '=' + v

                # update format
                res = res + server_params + ';\n'

            # set footer
            res = res + '}\n'

            # response
            return res

    except Exception as e:
        log.error('An error occurred while processing data in Nginx format: {}'.format(e))

    return res


def nginx_dhparam_validation(name, dhparam):
    try:

        # b64decode
        dhparam = base64.b64decode(dhparam)

        # file validation
        if load_pem_parameters(dhparam):
            return True, ''

    except Exception as e:
        err_msg = 'An error occurred while validation the DH parameters {}: {}'.format(name, e)
        return False, err_msg


def nginx_certificate_validation(name, crt, key):
    try:

        # b64decode
        crt = base64.b64decode(crt)
        key = base64.b64decode(key)

        # Convert key and cert to obj
        crt = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, crt)
        key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)

        ct = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_METHOD)
        ct.use_certificate(crt)
        ct.use_privatekey(key)

        # checking the key matches the certificate
        ct.check_privatekey()
        return True, ''

    except Exception as e:
        err_msg = 'An error occurred while validation the certificate {}: {}'.format(name, e)
        return False, err_msg


def nginx_value_extend_validation(waf_id, k, v):
    try:

        # init
        err_prefix = 'An error occurred during extended validation ({}/{})'.format(k, v)

        ##
        # Value validation
        ##

        # listen address[:port] [default_server] [ssl] [http2 | spdy] [proxy_protocol]
        if k == 'listen':
            
            # split to list
            v_list = v.split()

            # list validation
            if len(v_list) < 1:
                return False, '{}: address or/and port is not set'.format(err_prefix)

            ##
            # Address validation
            ##

            # init
            host = ''
            port = 0

            # host and port | host in ipv6
            if ':' in v_list[0]:

                # unix
                if v_list[0].startswith('unix:'):
                    return False, '{}: UNIX-domain sockets is not allowed'.format(err_prefix)
                
                # if ipv6
                elif '[' in v_list[0]:
                    addr_list = [x for x in re.split(r'(\[.+])', v_list[0]) if x]
                    host = addr_list[0]
                    if len(addr_list) > 1:
                        port = int(addr_list[1].split(':')[1])
                
                # ipv4
                else:
                    host = v_list[0].split(':')[0]
                    port = int(v_list[0].split(':')[1])

            # host or port only
            else:
                port = int(v_list[0]) if isinstance(v_list[0], int) else host
                
            # host validation
            if host:
                if host != '*' and not ip_domain_validation(host):
                    return False, '{}: incorrect IP address'.format(err_prefix)
            
            # port validation
            if port:
                if not isinstance(port, int) or 65535 < port < 1:
                    return False, '{}: incorrect port'.format(err_prefix)                

            # mutual exclusion validation
            if 'http2' in v and 'spdy' in v:
                return False, '{}: use of mutually exclusive parameters'.format(err_prefix)

        # ip(6)|ip(6)/cidr
        elif k == 'set_real_ip_from':
            for i in v:            
                if not ip_validation(i):
                    return False, '{}: incorrect IP address'.format(err_prefix)

        # X-Real-IP|X-Forwarded-For|proxy_protocol
        elif k == 'real_ip_header':
            if v.lower() not in ['x-real-ip', 'x-forwarded-for', 'proxy_protocol']:
                return False, '{}'.format(err_prefix)
                
        # zone=name [burst=number] [nodelay | delay=number]
        elif k == 'limit_req':

            # zone validation
            sep = 'zone='
            r = [x for x in v.split() if x.startswith(sep)]
            zone = r[0].split(sep)[1] if r else False
            if len(r) > 1 or not zone:
                return False, '{}: zone is not set'.format(k, v)

            # burst validation
            params = ['burst=', 'delay=']
            for param in params:
                sep = param
                r = [x for x in v.split() if x.startswith(sep)]
                if r:
                    p = int(r[0].split(sep)[1]) if r else False
                    if len(r) > 1 or not isinstance(p, int):
                        return False, '{}'.format(err_prefix)
                
            # nodelay validation
            sep = 'nodelay'
            r = [x for x in v.split() if x.startswith(sep)]
            if r:
                if len(r) > 1 or r[0] != sep:
                    return False, '{}'.format(err_prefix)
            
            # mutual exclusion validation
            if [x for x in v.split() if x.startswith('nodelay')] and [x for x in v.split() if x.startswith('delay')]:
                return False, '{}: use of mutually exclusive parameters'.format(err_prefix)

            ##
            # Search data in the database
            ##

            # init connection to the database
            connect, cur = db_conn()
            if not connect or not cur:
                return False, 'An error occurred while connection to the database'

            # search zone name
            query = """ SELECT limit_req_zone FROM nginx_global WHERE waf_id = %s """
            cur.execute(query, (waf_id,))
            res = cur.fetchall()
            connect.close()

            # processing result
            if not res or not [x for x in res[0][0].split() if x.startswith('zone=') and x.split('zone=')[1].split(':')[0] == zone]:
                return False, '{}: limit_req_zone "{}" is not found'.format(err_prefix, zone)

        # key zone=name:size rate=rate 
        # $binary_remote_addr zone=my_zone:10m rate=1r/s
        elif k == 'limit_req_zone':

            # processing the list
            for item in v:
            
                # key validation
                r = [x for x in item.split() if x.startswith('$')]
                if len(r) != 1 or r[0] not in ['$binary_remote_addr', '$remote_addr']:
                    return False, '{}: incorrect key data'.format(err_prefix)
                
                # zone validation
                sep = 'zone='
                r = [x for x in item.split() if x.startswith(sep)]
                if len(r) != 1:
                    return False, '{}: incorrect zone data'.format(err_prefix)
                if not re.match(r'^[\w\-.]+$', r[0].split(':')[0].split(sep)[1]):
                    return False, '{}: incorrect zone name'.format(err_prefix)
                if not re.match(r'^\d+[bkm]$', r[0].split(':')[1]):
                    return False, '{}: incorrect zone size'.format(err_prefix)

                # rate validation
                sep = 'rate='
                r = [x for x in item.split() if x.startswith(sep)]
                if len(r) != 1 or not re.match(r'^\d+r/[sm]$', r[0].split('=')[1]):
                    return False, '{}: incorrect rate data'.format(err_prefix)

        # TLSv1 TLSv1.1 TLSv1.2
        elif k == 'ssl_protocols':
            
            # init
            v_list = v.lower().split()
            allowed = ['tlsv1', 'tlsv1.1', 'tlsv1.2', 'tlsv1.3']

            # ssl protocols validation
            if [x for x in v_list if x not in allowed]:
                return False, '{}'.format(err_prefix)

        # 1.0|1.1
        elif k == 'proxy_http_version':
            if v not in [1.0, 1.1]:
                return False, '{}'.format(err_prefix)

        # example.com|example.com:80|[::]:443
        elif k == 'server':
           
            # init
            host = ''
            port = 0

            # host and port | host in ipv6
            if ':' in v:

                # unix
                if v.startswith('unix:'):
                    return False, '{}: UNIX-domain sockets is not allowed'.format(err_prefix)
                
                # if ipv6
                elif '[' in v:
                    addr_list = [x for x in re.split(r'(\[.+])', v) if x]
                    host = addr_list[0]
                    if len(addr_list) > 1:
                        port = int(addr_list[1].split(':')[1])
                
                # ipv4
                else:
                    host = v.split(':')[0]
                    port = int(v.split(':')[1])

            # host or port only
            else:
                port = int(v) if isinstance(v, int) else host
                
            # host validation
            if host:
                if host != '*' and not ip_domain_validation(host):
                    return False, '{}: incorrect IP address'.format(err_prefix)
            
            # port validation
            if port:
                if not isinstance(port, int) or 65535 < port < 1:
                    return False, '{}: incorrect port'.format(err_prefix)    
        
        # hash|ip_hash|least_conn|random
        elif k == 'balancing_method':
            if v not in ['hash', 'ip_hash', 'least_conn', 'random']:
                return False, '{}'.format(err_prefix)
        
        # error_page code ... [=[response]] uri;
        elif k == 'error_page':
            
            # split to list
            v = v.split()

            # list processing
            for i in v:

                # code validation
                if isinstance(i, int) and 99 < i < 600:
                    continue

                # response validation
                elif i.startswith('=') and i != '=':
                    
                    # extract response
                    i = i.split('=')[1]
                    
                    # validate the location
                    if isinstance(i, int) and 99 < i < 600:
                        continue

                # uri validation
                elif i.startswith('@'):

                    ##
                    # Search data in the database
                    ##

                    # init connection to the database
                    connect, cur = db_conn()
                    if not connect or not cur:
                        return False, 'An error occurred while connection to the database'

                    # search location
                    query = """ SELECT locations FROM nginx_vhost WHERE waf_id = %s """
                    cur.execute(query, (waf_id,))
                    res = cur.fetchall()
                    connect.close()

                    # processing result
                    if not res or not [x for x in res[0][0] if x['location'] == i]:
                        return False, '{}: location "{}" is not found'.format(err_prefix, i)

        # http://proxy_pass_upstream
        # memcached_pass_upstream
        elif k in ['proxy_pass', 'memcached_pass']:

            # proxy_pass preprocessing
            if k == 'proxy_pass':
                
                # should be starting with schema
                if not v.startswith('http://') and not v.startswith('https://'):
                    return False, '{}'.format(err_prefix)
                
                # extract upstream name
                i = v.split('://')[1]
            
            else:

                # extract upstream name
                i = v
                
            ##
            # Search data in the database
            ##

            # init connection to the database
            connect, cur = db_conn()
            if not connect or not cur:
                return False, 'An error occurred while connection to the database'

            # search upstream name
            query = """ SELECT COUNT(*) FROM nginx_upstream WHERE (waf_id = %s AND name = %s) """
            cur.execute(query, (waf_id, i,))
            res = cur.fetchone()
            connect.close()

            # processing result
            if not res or res[0] < 1:
                return False, '{}: upstream "{}" is not found'.format(err_prefix, i)

        # pass_type validation
        elif k == 'pass_type':
            if v not in ['root', 'alias', 'proxy_pass', 'memcached_pass']:
                return False, '{}'.format(err_prefix)

        # resolver address ... [valid=time] [ipv4=on|off] [ipv6=on|off];
        elif k == 'resolver':
 
            # convert to list
            v = v.split()

            # skip empty list
            if not v:
                return True, ''
            
            ##
            # Address validation
            ##

            # extract address and remove it from the list
            addr_str = v[0]
            del v[0]
            
            # init
            host = ''
            port = 0

            # host and port | host in ipv6
            if ':' in v:

                # unix
                if addr_str.startswith('unix:'):
                    return False, '{}: UNIX-domain sockets is not allowed'.format(err_prefix)
                
                # if ipv6
                elif '[' in addr_str:
                    addr_list = [x for x in re.split(r'(\[.+])', addr_str) if x]
                    host = addr_list[0]
                    if len(addr_list) > 1:
                        port = int(addr_list[1].split(':')[1])
                
                # ipv4
                else:
                    host = addr_str.split(':')[0]
                    port = int(addr_str.split(':')[1])

            # host or port only
            else:
                port = int(addr_str) if isinstance(addr_str, int) else host
                
            # host validation
            if host:
                if host != '*' and not ip_domain_validation(host):
                    return False, '{}: incorrect IP address'.format(err_prefix)
            
            # port validation
            if port:
                if not isinstance(port, int) or 65535 < port < 1:
                    return False, '{}: incorrect port'.format(err_prefix) 
                
            ##
            # Other params validation
            ##

            for i in v:

                # check valid
                if i.startswith('valid'):
                    i = i.split('=')[1]
                    if not isinstance(i, int):
                        return False, '{}'.format(err_prefix)

                # check ipv4/ipv6
                elif i.startswith('ipv'):
                    i = i.split('=')[1]
                    if i and i not in ['on', 'off']:
                        return False, '{}'.format(err_prefix)

                # check IP/domain
                else:

                    # IPv6 normalization
                    if i.startswith('['):
                        i = i.replace('[', '')
                        i = i.replace(']', '')

                    # IP/domain validation
                    if not ip_domain_validation(i):
                        return False, '{}: incorrect address'.format(err_prefix)

        # on/off
        elif k in ['autoindex', 'proxy_intercept_error']:
            if v not in ['on', 'off']:
                return False, '{}'.format(err_prefix)

        # certificate/dhparam/upstream
        if k in ['certificate_name', 'dhparam_name']:

            ##
            # Search data in the database
            ##

            # init connection to the database
            connect, cur = db_conn()
            if not connect or not cur:
                return False, 'An error occurred while connection to the database'

            # set type and table name
            nginx_type = 'certificate' if k == 'certificate_name' else 'dhparam'
            table = 'nginx_' + nginx_type

            # search upstream name
            query = """ SELECT COUNT(*) FROM {} WHERE (waf_id = %s AND name = %s) """.format(table)
            cur.execute(query, (waf_id, v,))
            res = cur.fetchone()
            connect.close()

            # processing result
            if not res or res[0] < 1:
                return False, '{}: {} "{}" is not found'.format(err_prefix, nginx_type, v)
            
        # finish processing
        return True, ''

    except Exception as e:
        
        if k is None:
            err_msg = 'An error occurred during extended validation "{}" value: {}'.format(v, e)
        else:
            err_msg = 'An error occurred during extended validation "{}" value of "{}" key: {}'.format(v, k, e)

        log.error(err_msg)
        return False, err_msg


def nginx_value_basic_validation(k, v):
    try:

        # value of 'name' type
        if k in nginx_name:
            if not re.match(r'^[\w\-.]+$', v):
                return False, 'The value {} of parameter {} contains an invalid characters'.format(v, k)
            else:
                return True, ''

        # value is a str
        elif k in nginx_type_str:
            if v is None:
                return True, ''
            elif not isinstance(v, str):
                return False, 'The value {} of parameter {} is not a string or not a null'.format(v, k)
            else:
                return True, ''
        
        # value is a b64
        elif k in nginx_type_b64:
            if v is None:
                return True, ''
            elif not is_base64(v):
                return False, 'The value {} of parameter {} is not in base64 or not a null'.format(v, k)
            else:
                return True, ''

        # value is a int
        elif k in nginx_type_int:
            if v is None:
                return True, ''
            elif not isinstance(v, int):
                return False, 'The value {} of parameter {} is not an integer or not a null'.format(v, k)
            else:
                return True, ''

        # value is a float
        elif k in nginx_type_float:
            if v is None:
                return True, ''
            elif not isinstance(v, float):
                return False, 'The value {} of parameter {} is not a float or not a null'.format(v, k)
            else:
                return True, ''

        # value is a bool
        elif k in nginx_type_bool:
            if not isinstance(v, bool):
                return False, 'The value {} of parameter {} is not a boolean'.format(v, k)
            else:
                return True, ''

        # value is a list
        elif k in nginx_type_list:
            if v is None:
                return True, ''
            else:
                if not isinstance(v, list):
                    return False, 'The value {} of parameter {} is not a list or not a null'.format(v, k)
                else:
                    return True, ''

        # value is a dict
        elif k in nginx_type_dict:
            if v is None:
                return True, ''
            else:
                if not isinstance(v, dict):
                    return False, 'The value {} of parameter {} is not a dictionary or not a null'.format(v, k)
                else:
                    return True, ''

        # value is a list of dict
        elif k in nginx_type_lod:
            if v is None:
                return True, ''
            else:
                for i in v:
                    if not isinstance(i, dict):
                        return False, 'The value {} of parameter {} is not a list of dictionaries or not a null'.format(v, k)
                return True, ''

    except Exception as e:
        
        if k is None:
            err_msg = 'An error occurred during validation "{}" value: {}'.format(v, e)
        else:
            err_msg = 'An error occurred during validation "{}" value of "{}" key: {}'.format(v, k, e)

        log.error(err_msg)
        return False, err_msg

    # response
    err_msg = 'Unexpected scenario in nginx_value_basic_validation ({}/{})'.format(k, v)
    return False, err_msg


def nginx_validation(waf_id, data, k, v):
    try:
        
        ##
        # Basic validation
        ##

        r = nginx_value_basic_validation(k, v)
        if not r[0]:
            return r[0], r[1]
        
        ##
        # Extended validation
        ##
        
        # skip None (None processing in basic validation)
        if v is None:
            return True, ''
        
        # list of dictionary
        if k in nginx_type_lod:
            # processing list
            for vi in v:                
                # processing dict
                for kvi, vvi in vi.items():
                    r = nginx_value_extend_validation(waf_id, kvi, vvi)
                    if not r[0]:
                        return r[0], r[1]

        # other types (e.g., str/int/list/dict)
        else:
            r = nginx_value_extend_validation(waf_id, k, v)
            if not r[0]:
                return r[0], r[1]

        ##
        # Certificate validation
        # (certificate and private key should be in data)
        ##
        
        if k == 'certificate':
            if data['certificate'] and data['certificate_key']:
                r = nginx_certificate_validation(data['name'], data['certificate'], data['certificate_key'])
                if not r[0]:
                    return r[0], r[1]
            else:
                err_msg = 'An error occurred while updating the certificate {}: certificate and private key are missing'.format(data['name'])
                log.error(err_msg)
                return False, err_msg
            
        ##
        # DH parameters validation
        ##

        if k == 'dhparam':
            r = nginx_dhparam_validation(data['name'], v)
            if not r[0]:
                return r[0], r[1]

    except Exception as e:
        err_msg = 'An error occurred during Nginx validation: {}/{}: {}'.format(k, v, e)
        log.error(err_msg)
        return False, err_msg

    return True, ''
