#!/usr/bin/env python3

import hashlib
import json

from psycopg2.extras import RealDictCursor
from pymemcache.client.base import Client

from core import base_api_uri
from core import db_conn
from core import log
from core import memcached_setting

##
# URLs
##

set_openapi_schema_uri = base_api_uri + '/set_openapi_schema'
get_openapi_schemas_uri = base_api_uri + '/get_openapi_schemas'
get_openapi_schemas_hash_uri = base_api_uri + '/get_openapi_schemas_hash'
get_openapi_url_uri = base_api_uri + '/get_openapi_url'
set_openapi_url_uri = base_api_uri + '/set_openapi_url'
get_openapi_url_hash_uri = base_api_uri + '/get_openapi_url_hash'
set_openapi_schema_paths_uri = base_api_uri + '/set_openapi_schema_paths'
get_openapi_schema_paths_uri = base_api_uri + '/get_openapi_schema_paths'
get_openapi_exclude_paths_uri = base_api_uri + '/get_openapi_exclude_paths'
set_openapi_exclude_paths_uri = base_api_uri + '/set_openapi_exclude_paths'
get_openapi_exclude_paths_hash_uri = base_api_uri + '/get_openapi_exclude_paths_hash'


def schema_name_after_change(old_name, new_name, waf_id):

    # init connection to the database
    connect, cur = db_conn()
    cur = connect.cursor(cursor_factory=RealDictCursor)
    if not connect or not cur:
        return False

    # update hash for schemas
    r = set_openapi_hash('openapi_schemas', waf_id)
    if r is False:
        connect.close()
        return False

    # update url
    query = """ SELECT convert_from(url,'UTF8') AS url FROM openapi_url WHERE waf_id = %s """
    cur.execute(query, (waf_id,))
    rows = cur.fetchone()

    if rows:

        try:
            url = json.loads(rows.get('url'))
        except json.JSONDecodeError:
            url = rows.strip('[]').split(',')

        url = [i.replace(old_name, new_name, 1) for i in url]

        try:
            url = json.dumps(url)
        except json.JSONDecodeError:
            url = str(url)

        query = """ UPDATE openapi_url SET url = %s, timestamp = CURRENT_TIMESTAMP WHERE (waf_id = %s) """
        cur.execute(query, (url, waf_id))
        connect.commit()

        r = set_openapi_hash('openapi_url', waf_id)
        if r is False:
            connect.close()
            return False

    # update path
    query = """ SELECT id, url, schema_name FROM openapi_schema_paths WHERE waf_id = %s AND schema_name = %s"""
    cur.execute(query, (waf_id, old_name))
    rows = cur.fetchall()

    for i in rows:
        url = i.get('url', '').replace(old_name, new_name, 1)
        query = """ UPDATE openapi_schema_paths SET url = %s, schema_name = %s, timestamp = CURRENT_TIMESTAMP WHERE (id = %s) """
        cur.execute(query, (url, new_name, i.get('id')))
    connect.commit()

    # update exclude path
    query = """ SELECT convert_from(exclude_paths,'UTF8') AS exclude_paths FROM openapi_exclude_paths WHERE waf_id = %s """
    cur.execute(query, (waf_id,))
    rows = cur.fetchone()

    if rows:

        try:
            exclude_paths = json.loads(rows.get('exclude_paths'))
        except json.JSONDecodeError:
            exclude_paths = rows.strip('[]').split(',')

        exclude_paths = [i.replace(old_name, new_name, 1) for i in exclude_paths]

        try:
            exclude_paths = json.dumps(exclude_paths)
        except json.JSONDecodeError:
            exclude_paths = str(exclude_paths)

        query = """ UPDATE openapi_exclude_paths SET exclude_paths = %s, timestamp = CURRENT_TIMESTAMP WHERE (waf_id = %s) """
        cur.execute(query, (exclude_paths, waf_id))
        connect.commit()

    r = set_openapi_hash('openapi_exclude_paths', waf_id)
    if r is False:
        connect.close()
        return False

    # close database connection
    connect.close()

    return True


def set_openapi_hash(table, waf_id):

    # init memcached
    client = Client(memcached_setting)

    # init connection to the database
    connect, cur = db_conn()
    if not connect or not cur:
        return False

    if table == 'openapi_schemas':

        field = 'schema'
        memc_key = 'hash_schemas_' + waf_id
        query = """
            SELECT
                timestamp, convert_from(schema,'UTF8') AS schema
            FROM
                openapi_schemas
            WHERE
                waf_id = %s
            ORDER BY
                timestamp
            DESC
        """

    elif table == 'openapi_url':

        memc_key = 'hash_url_' + waf_id
        field = 'url'
        query = """
            SELECT
                timestamp, convert_from(url,'UTF8') AS url
            FROM
                openapi_url
            WHERE
                waf_id = %s
        """

    elif table == 'openapi_exclude_paths':

        memc_key = 'hash_exclude_path_' + waf_id
        field = 'exclude_paths'
        query = """
            SELECT
                timestamp, convert_from(exclude_paths,'UTF8') AS exclude_paths
            FROM
                openapi_exclude_paths
            WHERE
                waf_id = %s
        """

    else:
        log.error('An error occurred while generate hash: table {} is not valid'.format(table))
        return False

    cur.execute(query, (waf_id,))
    rows = cur.fetchall()
    connect.close()

    h = waf_id
    if field == 'schema':
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)

        for i in rows:
            h += str(i[1])
    else:
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)
            try:
                rows = json.loads(rows[0][1])
            except json.JSONDecodeError:
                rows = rows[0][1].strip('[]').split(',')

        for i in rows:
            h += str(i)

    if rows:
        hash_elms = hashlib.sha256(h.encode()).hexdigest()
        client.set(memc_key, hash_elms)
    else:
        client.delete(memc_key)

    return True
