#!/usr/bin/python3

import logging

from time import time, sleep

from core import db_conn
from core import get_geoip_cc
from core import get_rldesc
from core import ro_mode
from core import waf_id_ce

from ec import event_eol


##
# Misc.
##

log_file = '/var/log/uwsgi/nw-api/rldscupd-logging.log'

##
# Log settings
##

log = logging.getLogger('main')
log.setLevel(logging.INFO)
formatter = logging.Formatter('[%(asctime)s] {%(name)s:%(lineno)d} %(levelname)s - %(message)s', '%m-%d %H:%M:%S')
file_handler = logging.FileHandler(log_file)
file_handler.setFormatter(formatter)
log.addHandler(file_handler)
log.info('START')


def db_update():
    try:

        # RW DB type only
        if ro_mode:
            return

        # connect to the DB
        connect, cur = db_conn()
        if not connect or not cur:
            return

        ##
        # Processing rule description in attack
        ##

        sql = """
            SELECT DISTINCT
                rule_id
            FROM
                attack
            WHERE
                description IS NULL OR description = '' OR description = '-'
        """
        cur.execute(sql,)
        rows = cur.fetchall()

        if len(rows):
            for i in rows:

                # init
                rule_id = str(i[0])

                if rule_id == '0':
                    continue

                # receive rule description if rule id not 0
                rldesc = get_rldesc(rule_id)
                log.info('Updated description for RuleID {}: {}'.format(rule_id, rldesc))

                # if rule description is not empty
                if rldesc:

                    # update description
                    sql = """ UPDATE attack SET description = %s WHERE rule_id = %s """
                    ar = (rldesc, rule_id,)
                    cur.execute(sql, ar)
                    connect.commit()

        ##
        # Processing CC in attack
        ##

        tables = ['attack', 'mt_main']

        for t in tables:
            try:
        
                sql = """
                    SELECT DISTINCT
                        ip, waf_id
                    FROM
                        {}
                    WHERE
                        (
                            cc IS NULL
                                AND
                            waf_id != %s
                                AND
                            waf_id != ''
                        )
                    LIMIT 1000
                """.format(t)
                cur.execute(sql, (waf_id_ce,))
                rows = cur.fetchall()

                # skip empty
                if len(rows) < 1:
                    continue

                ##
                # Updating CC
                ##
                
                for i in rows:

                    # extrack IP and WAF ID
                    ip = str(i[0]).strip()
                    waf_id = str(i[1]).strip()

                    # get CC from remote server
                    cc = get_geoip_cc(ip, waf_id)

                    # update CountryCode
                    if cc:

                        # logging
                        if cc:
                            log.info('Updating CountryCode for {} ({})'.format(ip, cc))
                        else:
                            log.info('Updating CountryCode for {}'.format(ip))

                        # update
                        sql = """ UPDATE {} SET cc = %s WHERE ip = %s """.format(t)
                        ar = (cc, ip,)
                        cur.execute(sql, ar)
                        connect.commit()

            except Exception as e:
                log.error('An error occurred while updating CountryCode in {}: {}'.format(t, e))

        ##
        # Delete old events
        ##

        try:
        
            # Set a timestamp, less than which all events will be deleted
            eol_ts = int(time()) - event_eol

            # Delete an events
            sql = """ DELETE FROM events WHERE timestamp < %s """
            cur.execute(sql, (eol_ts,))
            connect.commit()

        except Exception as e:
            log.error('An error occurred while deleting old events: {}'.format(e))

        # close connection to the DB
        connect.close()

    except Exception as e:
        log.error('An error occurred while processing data: {}'.format(e))


if __name__ == '__main__':
    while True:
        db_update()
        sleep(60)
