#!/usr/bin/env python3

import ast
import base64
import hashlib
import re
import validators

from datetime import datetime
from distutils.util import strtobool
from netaddr import IPNetwork, IPRange

from core import base_api_uri
from core import db_conn
from core import log

##
# URLs
##

get_vhosts_list_uri = base_api_uri + '/get_vhosts_list'
set_vhosts_list_uri = base_api_uri + '/set_vhosts_list'

get_mlc_settings_hash_uri = base_api_uri + '/get_mlc_settings_hash'
get_mlc_settings_uri = base_api_uri + '/get_mlc_settings'
set_mlc_settings_uri = base_api_uri + '/set_mlc_settings'

get_dyn_settings_hash_uri = base_api_uri + '/get_dyn_settings_hash'
get_dyn_settings_uri = base_api_uri + '/get_dyn_settings'
set_dyn_settings_uri = base_api_uri + '/set_dyn_settings'

get_dyn_wl_hash_uri = base_api_uri + '/get_dyn_wl_hash'
get_dyn_wl_uri = base_api_uri + '/get_dyn_wl'
set_dyn_wl_uri = base_api_uri + '/set_dyn_wl'

get_dyn_erl_hash_uri = base_api_uri + '/get_dyn_erl_hash'
get_dyn_erl_uri = base_api_uri + '/get_dyn_erl'
set_dyn_erl_uri = base_api_uri + '/set_dyn_erl'

get_dyn_bl_hash_uri = base_api_uri + '/get_dyn_bl_hash'
get_dyn_bl_uri = base_api_uri + '/get_dyn_bl'
set_dyn_bl_uri = base_api_uri + '/set_dyn_bl'

get_dyn_sync_ban_uri = base_api_uri + '/get_dyn_sync_ban'
set_dyn_sync_ban_uri = base_api_uri + '/set_dyn_sync_ban'
del_dyn_sync_ban_uri = base_api_uri + '/del_dyn_sync_ban'

##
# MLC settings keys
##

# set forbidden columns by tarif_id
mlc_be_cols = ['main__ai_extra', 'openapi__vhosts_list']

# set columns
mlc_settings_cols = [
    'main__ai_extra',
    'ddos__enable',
    'ddos__wl_ip',
    'ddos__wl_url',
    'ddos__interval',
    'ddos__latest_only',
    'ddos__send_possible',
    'brute__enable',
    'brute__brute_detect',
    'brute__flood_detect',
    'brute__wl_host',
    'brute__interval',
    'brute__max_val',
    'brute__latest_only',
    'brute__send_possible',
    'openapi__vhosts_list'
]

# set types
mlc_settings_type_ip = ['ddos__wl_ip']
mlc_settings_type_domain = ['brute__wl_host', 'openapi__vhosts_list']
mlc_settings_type_url = ['ddos__wl_url', 'brute__brute_detect', 'brute__flood_detect']
mlc_settings_type_int = ['ddos__interval', 'brute__interval', 'brute__max_val']
mlc_settings_type_bool = [
    'main__ai_extra', 'ddos__enable', 'ddos__latest_only', 'ddos__send_possible',
    'brute__enable', 'brute__latest_only', 'brute__send_possible'
]

##
# Dyn settings keys
##

# set columns
dyn_settings_cols = [
    'wl',
    'rl',
    'nwaf_ai_extra_host_lm',
    'nwaf_ai_extra_host_wl',
    'nwaf_bf_detect_host_lm',
    'nwaf_body_exclude',
    'nwaf_b64_decode_disable',
    'nwaf_ddos_detect_host_lm',
    'nwaf_geoip_mla_disable',
    'nwaf_url_wl',
    'nwaf_host_lm',
    'nwaf_host_wl',
    'nwaf_ip_lm',
    'nwaf_ip_wl',
    'nwaf_limit',
    'nwaf_mla_host_lm',
    'nwaf_post_body_exclude',
    'nwaf_put_body_exclude',
    'nwaf_rmq_host_exclude',
    'nwaf_openapi_ip_wl',
    'nwaf_openapi_ip_lm',
    'nwaf_openapi_url_wl',
    'nwaf_openapi_url_lm',
    'nwaf_rfc_violation_wl',
    'nwaf_rfc_violation_lm'
]

##
# Set types
##

dyn_settings_type_limit = [
    'nwaf_limit'
]

dyn_settings_type_zone = [
    'nwaf_b64_decode_disable'
]

dyn_settings_type_nwaf_ip = [
    'nwaf_ip_wl', 'nwaf_ip_lm',
    'nwaf_openapi_ip_wl', 'nwaf_openapi_ip_lm'
]

dyn_settings_type_url = [
    'nwaf_body_exclude',
    'nwaf_url_wl',
    'nwaf_openapi_url_wl', 'nwaf_openapi_url_lm',
    'nwaf_rfc_violation_wl', 'nwaf_rfc_violation_lm',
]

dyn_settings_type_domain = [
    'nwaf_ddos_detect_host_lm',
    'nwaf_host_wl', 'nwaf_host_lm',
    'nwaf_ai_extra_host_wl', 'nwaf_ai_extra_host_lm',
    'nwaf_bf_detect_host_lm',
    'nwaf_geoip_mla_disable',
    'nwaf_mla_host_lm',
    'nwaf_put_body_exclude'
    'nwaf_post_body_exclude'
    'nwaf_rmq_host_exclude',
]

##
# Dyn WL keys
##

# set columns
dyn_wl_cols = [
    'id',
    'domain',
    'mz',
    'extension',
    'rl_id'
]

##
# Dyn BL keys
##

# set columns
dyn_bl_cols = ['bl']

##
# Dyn ERL keys
##

# set types
dyn_erl_type_boolean = ['api', 'lm', 'noban', 'no_cookie']
dyn_erl_type_data = ['ip', 'country', 'domain', 'url', 'args', 'body', 'cookie', 'ua', 'referer', 'other_headers']

# set columns
dyn_erl_cols = dyn_erl_type_boolean + dyn_erl_type_data

##
# Dyn SyncBan keys
##

dyn_sync_ban_dict = ['ban_data']
dyn_sync_ban_int = ['timestamp']

##
# Set list of allowed command
##

settings_cmd = ['set', 'del']  # (Dyn, MLC)
dyn_wl_erl_cmd = ['add', 'upd', 'del']  # Dyn WL/ERL/BL


def base64_decode(agent_ip, k, v, s):

    # convert to binary
    try:

        # validation
        status = base64.b64encode(base64.b64decode(s.encode())).decode('UTF-8', 'ignore') == s

        # check status
        if not status:
            return False

        # return decoded
        return base64.b64decode(s).decode()

    except Exception:
        log.error(
            'An error occurred while B64 decoding value from {}: {}/{} ({})'
            .format(agent_ip, k, v, s)
        )
        return False


def dyn_ip_domain_validation(item_list):
    try:
        # domain validation
        for item in item_list:

            # flag api validation
            if item.startswith('api='):

                # check value
                if item != 'api=yes' and item != 'api=no':
                    return False

            # domain validation
            elif item.startswith('domain='):

                # extract values
                domain = item.split('domain=')[1]

                # processing '*'
                if domain == '*':
                    continue

                # remove '*.' or '.' for future domain validation
                domain = re.sub(r'^[*]?\.', '', domain)

                # check value
                if not validators.domain(domain):
                    return False

            # IP validation
            else:

                # IP range
                if '-' in item:
                    # processing IP range
                    IPRange(item.split('-')[0], item.split('-')[1])
                else:
                    # processing IPv4/IPv6/CIDR
                    IPNetwork(item)

    except Exception as e:
        log.error('An error occurred while processing data in dyn_ip_domain_validation: {}'.format(e))
        return False

    return True


def get_hash(agent_ip, url, t, waf_id):
    try:

        ##
        # Get ts
        ##

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            return None, 503

        query = """ SELECT modify FROM {} WHERE waf_id = %s """.format(t)
        cur.execute(query, (waf_id,))
        res = cur.fetchone()
        connect.close()

        # convert to timestamp
        if res:
            res = res[0].timestamp()
        else:
            res = 0

        # processing result
        if not res:
            resp = str({'status': 'none'})
            return resp, 200

        else:
            res = hashlib.md5(str(res).encode()).hexdigest()
            resp = {'status': 'success', 'hash': str(res)}
            return str(resp), 200

    except Exception as e:
        log.error('An error occurred in get_hash in {} from {}: {}'.format(agent_ip, url, e))

    # default response
    return None, 503


def set_hash(agent_ip, url, t, waf_id):
    try:

        ##
        # Update ts
        ##

        # init connection to the database
        connect, cur = db_conn()
        if not connect or not cur:
            return

        query = """ INSERT INTO {} (modify, waf_id) VALUES (%s, %s) ON CONFLICT (waf_id) DO UPDATE SET modify = %s """.format(t)
        cur.execute(query, (str(datetime.now()), waf_id, str(datetime.now()),))
        connect.commit()
        connect.close()

    except Exception as e:
        log.error('An error occurred in set_hash in {} from {}: {}'.format(agent_ip, url, e))


def mlc_value_validation(agent_ip, k, v):
    try:

        # remove spaces before and after
        if isinstance(v, str):
            v = v.strip()
            if not v:
                return True

        if k in mlc_settings_type_ip:

            # split to list
            v = v.split()

            # processing IPv4/IPv6/CIDR
            for item in v:
                IPNetwork(item)

        elif k in mlc_settings_type_domain:

            # processing list
            for item in v:

                # processing '*'
                if item == '*':
                    if k == 'brute__wl_host':
                        # '*' is not allowed for this option
                        return False
                    else:
                        continue

                # remove '*.' or '.' for future domain validation
                item = re.sub(r'^[*]?\.', '', item)

                # check value
                if not validators.domain(item):
                    return False

        elif k in mlc_settings_type_url:

            # processing list
            for item in v:

                # '*' is not allowed for this type
                if item == '*':
                    return False

                res = settings_type_url_validation(item)
                if res is False:
                    return False

        elif k in mlc_settings_type_int:

            # check value
            if not int(v) >= 0:
                return False

        elif k in mlc_settings_type_bool:

            # check value
            if type(v) is not bool:
                return False

    except Exception as e:
        log.error('An error occurred while validating MLC key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def dyn_value_validation(agent_ip, k, v):
    try:

        # remove spaces before and after
        v = v.strip()

        # skip empty
        if not v:
            return True

        # split values to list
        v_list = v.split(', ')

        if k in dyn_settings_type_nwaf_ip:

            # processing items
            for item in v_list:
                if not dyn_ip_domain_validation(item.split()):
                    return False

        elif k in dyn_settings_type_zone:

            # processing list
            for item in v_list:

                item_v_list = item.split()

                for item_v in item_v_list:

                    # zone validation
                    if item_v.startswith('z:'):

                        val = re.match(r'^z:([\w|]+)+$', item_v)

                        # check value
                        if not val:
                            return False

                        res = settings_type_zone_validation(val[1].split('|'))
                        if res is False:
                            return False

                    # url validation
                    else:

                        res = settings_type_url_validation(item_v)
                        if res is False:
                            return False

        elif k in dyn_settings_type_domain:

            # processing list
            for item in v_list:

                # processing '*'
                if item == '*':
                    continue

                # remove '*.' or '.' for future domain validation
                item = re.sub(r'^[*]?\.', '', item)

                # check value
                if not validators.domain(item):
                    return False

        elif k in dyn_settings_type_limit:

            """
            rate=5r/m (required)
            block_time=300 (required)
            domain=example.com (optional)
            """

            # processing list
            for item in v_list:

                # split item to list
                item_v_list = item.split()

                # check required parameters
                if 'rate=' not in ' '.join(item_v_list) or 'block_time=' not in ' '.join(item_v_list):
                    return False

                # processing list
                for item_v in item_v_list:

                    # if item is domain
                    if item_v.startswith('domain'):

                        # extract domain
                        domain = item_v.split('domain=')[1]

                        # processing '*'
                        if domain == '*':
                            continue

                        # remove '*.' or '.' for future domain validation
                        domain = re.sub(r'^[*]?\.', '', domain)

                        # check value
                        if not validators.domain(domain):
                            return False

                    # if item is rate
                    if item_v.startswith('rate'):

                        # extract value
                        rate_data = item_v.split('rate=')[1]

                        # extract rate and intensity
                        rate = rate_data.split('r/')[0]
                        rate_intensity = rate_data.split('r/')[1]

                        # rate rate intensity validation
                        if int(rate) < 0:
                            return False

                        # rate validation
                        if rate_intensity not in ['s', 'm', 'h']:
                            return False

                    # if item is block_time
                    if item_v.startswith('block_time'):

                        # extract value
                        block_time = item_v.split('block_time=')[1]

                        # value validation
                        if int(block_time) < 0:
                            return False

        elif k in dyn_settings_type_url:

            # processing list
            for item in v_list:

                res = settings_type_url_validation(item)
                if res is False:
                    return False

    except Exception as e:
        log.error('An error occurred while validating Dyn key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def settings_type_url_validation(item):
    try:

        # allow '*'
        if item == '*':
            return True

        # replace '*' to example.com for future URL validation
        item = re.sub(r'^\*/', 'example.com/', item)

        # remove '*.' or '.' for future URL validation
        item = re.sub(r'^[*]?\.', '', item)

        # add the schema for validation
        item = 'http://' + item

        # URL
        if not validators.url(item):
            return False
        else:
            return True

    except Exception as e:
        log.error('An error occurred while validating data {}: {}'.format(item, e))
        return False


def settings_type_zone_validation(item_list):
    try:

        list_zone = ['url', 'args', 'headers', 'body']

        # processing list
        for item in item_list:

            # check value
            if item.lower() not in list_zone:
                return False

        return True

    except Exception as e:
        log.error('An error occurred while processing data in settings_type_zone_validation: {}'.format(e))
        return False


def dyn_wl_value_validation(agent_ip, k, v):
    try:

        ##
        # RL/WL/LM validation
        ##

        # check domain
        if k == 'domain':

            # allow empty
            if not v:
                return True

            # init
            domain = v

            # processing '*'
            if domain == '*':
                return True

            # remove '*.' or '.' for future domain validation
            domain = re.sub(r'^[*]?\.', '', domain)

            # check value
            if not validators.domain(domain):
                return False

        # check Rule ID for WL/LM
        if k == 'rl_id':

            # check ID (only "*" or integer )
            if v != '*' and not re.search(r'^\d+$', v):
                return False

        # check zone
        if k == 'mz':

            # split zone to list
            z_list = v.split('|')

            # check zone
            for item in z_list:

                # only allowed symbols
                if not re.match(r'^[\w-]+$', item, re.I):
                    return False

        # check extension
        if k == 'extension':

            # split extension to list
            ext_list = v.split('|$')

            # check zone and extension
            for item in ext_list:

                # skip empty
                if not item:
                    continue

                # add "$" to extension if it removed while split
                if not item.startswith('$'):
                    item = '$' + item

                # only allowed symbols
                if not re.match(r'^\$[\w-]+(_x)?:', item, re.I):
                    return False

    except Exception as e:
        log.error('An error occurred while validating Dyn key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def dyn_bl_value_validation(agent_ip, k, v):
    try:

        if k == 'active':

            # processing values
            v = bool(strtobool(str(v)))

            # check value
            if type(v) is not bool:
                return False

        if k == 'bl':

            # represent values as list
            v_list = ast.literal_eval(v)

            # processing list
            for item in v_list:
                if not dyn_ip_domain_validation(item.split()):
                    return False

    except Exception as e:
        log.error(
            'An error occurred while validating BL key/value from {}: {}/{}: {}'
            .format(agent_ip, k, v, e)
        )
        return False

    return True


def dyn_erl_value_validation(agent_ip, k, v):
    try:

        if k in dyn_erl_type_boolean:

            # processing values
            v = bool(strtobool(str(v)))

            # check value
            if type(v) is not bool:
                return False

        if k == 'ip':

            # split values to list
            # and remove empty items
            v_list = list(filter(None, re.split(r'[!&|]', v)))

            # processing list
            for item in v_list:

                # if IP range
                if '-' in item:
                    # processing IP range
                    IPRange(item.split('-')[0], item.split('-')[1])
                else:
                    # processing IPv4/IPv6/CIDR
                    IPNetwork(item)

        elif k == 'country':

            # split values to list
            # and remove empty items
            v_list = list(filter(None, re.split(r'[!&|]', v)))

            # processing list
            for item in v_list:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                if not re.match(r'^[a-z]{2}$', item, re.I):
                    return False

        elif k == 'domain':

            # split values to list
            # and remove empty items
            v_list = list(filter(None, re.split(r'[!&|]', v)))

            # processing list
            for item in v_list:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                # processing '*'
                if item == '*':
                    continue

                # remove '*.' or '.' for future domain validation
                item = re.sub(r'^[*]?\.', '', item)

                # check value
                if not validators.domain(item):
                    return False

        elif k == 'url':

            # split values to list
            # and remove empty items
            v_list = list(filter(None, re.split(r'[!&|]', v)))

            # processing list
            for item in v_list:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                """
                URL must be without domain and schema (path only).
                Therefore, we specifically add parameters for verification.
                """

                # add schema and domain
                item = 'http://example.com/' + item

                # URL
                if not validators.url(item):
                    return False

        elif k in ['args', 'body', 'cookie', 'ua']:

            # split values to list
            # and remove empty items
            v_list = list(filter(None, re.split(r'[!&|]', v)))

            # processing list
            for item in v_list:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

        elif k == 'other_headers':

            # represent values as list
            v_list = ast.literal_eval(v)

            # processing list
            for item in v_list:

                # split value to list
                item_v_list = item.split()

                # processing list
                for item1 in item_v_list:

                    # split to k/v
                    item1_k = item1.split(':')[0].strip()
                    item1_v = item1.split(':')[1].strip()

                    # remove '!!' from the start of key
                    item1_k = re.sub(r'^!!', '', item1_k)

                    # init
                    s = item1_k

                    ##
                    # Base64 decode: key
                    ##

                    try:

                        # decode
                        s = base64_decode(agent_ip, k, v, s)

                        # decode validation
                        if s is False:
                            return False

                    except Exception as e:
                        log.error(
                            'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                            .format(agent_ip, k, v, s, e)
                        )
                        return False

                    ##
                    # Base64 decode: value
                    ##

                    # split values with "!", "&" and "|"
                    item1_v_list = list(filter(None, re.split(r'[!&|]', item1_v)))

                    # processing list
                    for item2 in item1_v_list:

                        # init
                        s = item2

                        # base64 decode
                        try:

                            # decode
                            s = base64_decode(agent_ip, k, v, s)

                            # decode validation
                            if s is False:
                                return False

                        except Exception as e:
                            log.error(
                                'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                                .format(agent_ip, k, v, s, e)
                            )
                            return False

    except Exception as e:
        log.error(
            'An error occurred while validating ERL key/value from {}: {}/{}: {}'
            .format(agent_ip, k, v, e)
        )
        return False

    return True


def dyn_sync_ban_value_validation(agent_ip, k, v):
    try:

        if k in dyn_sync_ban_dict:

            # load data as a dictionary
            v = ast.literal_eval(v)

            # processing empty values
            if not v:
                if not isinstance(v, list):
                    return False

            # processing non-empty values
            else:
                for i in range(len(v)):
                    if not isinstance(v[i], dict):
                        return False

        if k in dyn_sync_ban_int:

            # processing values
            int(v)

    except Exception as e:
        log.error(
            'An error occurred while validating DynSynBan key/value from {}: {}/{}: {}'
            .format(agent_ip, k, v, e)
        )
        return False

    return True
